## SCRIPT: annotate_nodes.py
## AUTHOR: Shyamalika Gopalan
## OVERVIEW: Adds metadata information to newly-generated nodes.
## REQUIRED ARGUMENTS:
##			-f, --file (path to tree sequence file to be annotated)
##			-i, --inf_id (infection ID that produced this tree sequence file)
##			-s, --simulation_tag (an additional tag that can be used to distinguish this simulation run - optional)
##
## USAGE EXAMPLE:
## $ python annotate_nodes.py --input_file ~/test/run/tree.ts --inf_id 123 --simulation_tag 1

# Import necessary libraries
import argparse as ap
import tskit
import os
import sys

## SUB-FUNCTIONS ----
## ------------------
# Drop provenance information
def drop_provenances(tree_seq):
	allTables = tree_seq.dump_tables()
	allTables.provenances.clear()
	new_tree_seq = allTables.tree_sequence()
	return new_tree_seq

# Create node id to SLiM id dictionary
def node_to_slim_dict(tree_seq, as_string=False):
	id_map = {}
	for i in range(tree_seq.num_nodes):
		if as_string:
			slim_id = str(tree_seq.nodes()[i].metadata.get('slim_id'))
		else:
			slim_id = tree_seq.nodes()[i].metadata.get('slim_id')
		id_map[i] = slim_id
	return(id_map)

## MAIN FUNCTION ----
## ------------------
def main(args):
	argp = ap.ArgumentParser(description = 'Adds metadata information to newly-generated nodes.')
	argp.add_argument('-f', '--file', type = str, help = 'Path to tree sequence file to be annotated.')
	argp.add_argument('-i', '--inf_id', type = str, help = 'Infection ID that produced this tree sequence file.')
	argp.add_argument('-s', '--simulation_tag', type = str, help = 'An additional tag that can be used to distinguish this simulation run - optional.', required=False)
	args = argp.parse_args(args)

	# Read in tree
	tree = drop_provenances(tskit.load(args.file))

	# Load SLiM_RM_map object, which maps each slim_id present in the tree to the inf_id that it arose in and, optionally, an additional simulation tag
	SLiM_RM_map = tree.metadata["SLiM"]["user_metadata"]["SLiM_RM_map"]
	if type(SLiM_RM_map)is list:
		SLiM_RM_map = SLiM_RM_map[0]
	#samples = tree.metadata["SLiM"]["user_metadata"]["samples"]

	# Create map from node id to SLiM id
	get_slim_id = node_to_slim_dict(tree)

	# Update SLiM_RM_map
	slim_ids = [str(get_slim_id.get(i)) for i in range(tree.num_nodes)]
	for i in slim_ids:
		if i not in SLiM_RM_map.keys():
			SLiM_RM_map[i] = {"inf_id":args.inf_id,"simulation_tag":args.simulation_tag}
		else:
			map_data = SLiM_RM_map.get(i)
			if type(map_data)is list:
				map_data = map_data[0]
			inf_id = map_data['inf_id']
			if type(inf_id)is list:
				inf_id = inf_id[0]
			simulation_tag = map_data['simulation_tag']
			if type(simulation_tag)is list:
				simulation_tag = simulation_tag[0]
			SLiM_RM_map[i] = {"inf_id":inf_id, "simulation_tag":simulation_tag}

	# Clean up the map (avoid carrying over unnecessary info)
	entries_to_remove = []
	for i in SLiM_RM_map.keys():
		if i not in slim_ids:
			entries_to_remove = entries_to_remove + [i]
	if len(entries_to_remove) > 0:
		print("Warning: some slim ids have been lost", flush = True)
		for i in entries_to_remove:
			del SLiM_RM_map[i]

	# Store the map in the top-level metadata
	metadata_schema = tree.metadata_schema.schema.copy()
	top_level_meta = tree.metadata
	top_level_meta["SLiM"]["user_metadata"]["SLiM_RM_map"] = SLiM_RM_map
	tree_tables = tree.dump_tables()
	tree_tables.metadata_schema = tskit.MetadataSchema(metadata_schema)
	tree_tables.metadata = top_level_meta

	# Re-make and write out the tree sequence
	tree = tree_tables.tree_sequence()
	tree.dump(os.path.splitext(args.file)[0] + '_annotated.ts')

if __name__ == "__main__":
	sys.exit(main(sys.argv[1:]))
