## SCRIPT: simplify_trees.py
## AUTHOR: Shyamalika Gopalan
## OVERVIEW: Filters out individuals that do not contribute to the history of the input tree sequences.
##
## USAGE EXAMPLE:
## $ python simplify_trees.py tree1.ts tree2.ts tree3.ts

# Import necessary libraries
import argparse as ap
import tskit
import numpy as np
import pandas as pd
import os
import sys

## SUB-FUNCTIONS ----
## ------------------
# Drop provenance information
def drop_provenances(tree_seq):
	allTables = tree_seq.dump_tables()
	allTables.provenances.clear()
	new_tree_seq = allTables.tree_sequence()
	return new_tree_seq

def remember_inds(tree_seq):
	allTables = tree_seq.dump_tables()
	allTables.individuals.clear()
	for i in range(tree_seq.num_individuals):
		flag = tree_seq.individuals_flags[i]
		if flag in [65536, 262144, 65536 + 262144]: # individual marked as 'alive', 'retained', or both
			flag = flag + 131072
		allTables.individuals.add_row(flags = flag, \
									  location = tree_seq.individuals_location[i],\
									  parents = tree_seq.tables.individuals[i].parents, \
									  metadata = tree_seq.tables.individuals[i].metadata)
	new_tree_seq = allTables.tree_sequence()
	return new_tree_seq

## MAIN FUNCTION ----
## ------------------
def main(args):
	argp = ap.ArgumentParser(description = 'Filters out permanently remembered individuals that do not contribute to the history of the input tree sequences.')
	argp.add_argument('tree_seqs', nargs='*', type = str, help = 'Paths to tree sequences to be processed.')
	args = argp.parse_args(args)

	# Load trees
	tree_seqs = {}
	slim_ids = {}
	for ts in args.tree_seqs:
		tree_seqs[ts] = drop_provenances(tskit.load(ts))
		slim_ids[ts] = [tree_seqs[ts].nodes()[i].metadata['slim_id'] for i in range(tree_seqs[ts].num_nodes)]

	# Determine simulation start time
	sim_start_day = np.unique([tree_seqs[i].metadata['SLiM']['user_metadata']['sim_start_time'] for i in tree_seqs.keys()])
	if len(sim_start_day) > 1:
		raise ValueError("Multiple unique sim_start_time values found.")
	else:
		sim_start_day = sim_start_day[0]

	# Simplify all trees on sample nodes
	simplified_trees = {}
	for ts in tree_seqs:
		tree = tree_seqs[ts]
		sample_nodes = tree.metadata['SLiM']['user_metadata']['samples']
		simplified_trees[ts] = tree.simplify(samples=[i for i in range(len(slim_ids[ts])) if slim_ids[ts][i] in sample_nodes], keep_unary = True, keep_input_roots = True, filter_sites = False, record_provenance = False)

	# Determine SLiM IDs for all nodes across all simplified trees
	all_slim_ids0 = pd.DataFrame(columns = ['slim_id', 'day_born'])
	for ts in simplified_trees:
		tree = simplified_trees[ts]
		slim_ids0 = [i.metadata['slim_id'] for i in tree.tables.nodes]
		day_born = tree.metadata['SLiM']['cycle'] - tree.nodes_time
		all_slim_ids0 = pd.concat([all_slim_ids0, pd.DataFrame({'slim_id': slim_ids0, 'day_born': day_born})], ignore_index=True)
	all_slim_ids0 = all_slim_ids0.drop_duplicates().sort_values(by='day_born')

	# Determine SLiM IDs for all nodes after all trees are simplified on shared nodes
	all_slim_ids = pd.DataFrame(columns = ['slim_id', 'day_born'])
	resimplified_trees = {}
	for ts in tree_seqs:
		tree = tree_seqs[ts]
		# If node was born during this simulation, keep only if on the list
		tree_time = tree.metadata['SLiM']['cycle']
		node_age_condition = tree.nodes_time >= (tree_time - sim_start_day)
		slim_id_info = pd.DataFrame({'slim_id': slim_ids[ts], 'day_born': tree_time - tree.nodes_time})
		slim_id_condition = np.array(slim_id_info.apply(tuple, axis=1).isin(all_slim_ids0.apply(tuple, axis=1)))
		nodes_to_keep = list(np.where(node_age_condition | slim_id_condition)[0])
		# Do a first round of simplification, keeping all necessary nodes
		tree_simplified0 = tree.simplify(samples=nodes_to_keep, keep_unary = True, keep_input_roots = True, filter_individuals = False, filter_sites = False, filter_populations = False, record_provenance = False)
		# If any individual's second genome is not in nodes_to_keep, add it
		slim_ids_sim = [j for j in [tree_simplified0.nodes()[i].metadata['slim_id'] for i in range(tree_simplified0.num_nodes)] if j % 2 == 0]
		slim_ids_sim = np.sort(slim_ids_sim + [i+1 for i in slim_ids_sim])
		nodes_to_add  = np.where(np.logical_and(np.isin(slim_ids[ts], slim_ids_sim), np.isin(range(tree.num_nodes), nodes_to_keep, invert = True)))[0]
		if len(nodes_to_add) > 0:
			nodes_to_keep = nodes_to_keep + list(nodes_to_add)
		# Simplify again
		resimplified_trees[ts] = tree.simplify(samples=nodes_to_keep, keep_unary = True, keep_input_roots = True, filter_sites = False, filter_populations = False, record_provenance = False)
		# Store nodes
		slim_ids0 = [i.metadata['slim_id'] for i in resimplified_trees[ts].tables.nodes]
		day_born = resimplified_trees[ts].metadata['SLiM']['cycle'] - resimplified_trees[ts].nodes_time
		all_slim_ids = pd.concat([all_slim_ids, pd.DataFrame({'slim_id': slim_ids0, 'day_born': day_born})], ignore_index=True)
	all_slim_ids = all_slim_ids.drop_duplicates().sort_values(by='day_born')

	# Check if any nodes need to be re-labelled
	if len(all_slim_ids.slim_id) > len(np.unique(all_slim_ids.slim_id)):
		print("Warning: duplicate slim ids found", flush = True)

	# Set all individuals in the table to be permanently remembered
	for ts in tree_seqs:
		resimplified_trees[ts] = remember_inds(resimplified_trees[ts])

	# Write out tree
	for ts in tree_seqs:
		resimplified_trees[ts].dump(os.path.splitext(ts)[0] + '_simplified.ts')

if __name__ == "__main__":
	sys.exit(main(sys.argv[1:]))
